from .processor import Processor
from .state import State


class Node(object):
    def __init__(self, processor: Processor, childs: set, parent: str = "root"):
        self.processor = processor
        self.childs = childs
        self.parent = parent
        self.in_degree = processor.in_degree

    def check(self, state: State = None, conditions: dict = {}, validator=None):
        if not conditions:
            conditions = self.processor.conditions

        if isinstance(conditions, tuple):
            return state.get(conditions[0]) == conditions[1]

        if isinstance(conditions, dict) and not conditions:
            return True

        results = []

        if isinstance(conditions, dict):
            for key, value in conditions.items():
                if key == "not":
                    return not self.check(state, value)

                if key == "all":
                    return self.check(state, value, all)

                if key == "any":
                    return self.check(state, value, any)

                results.append(self.check(state, (key, value)))

        if isinstance(conditions, list):
            for condition in conditions:
                results.append(self.check(state, condition))

        if not validator:
            validator = all

        return validator(results)



import yaml

from .cortex import Cortex


class CortexManager:
    def __init__(self, config):
        self._cortices = {}

        for item in config:
            with open(item["path"], "r") as f:
                instance = Cortex()
                instance.build(item["name"], yaml.safe_load(f))

                self._cortices[item["name"]] = instance

    def get(self, name):
        return self._cortices.get(name, None)



from .class_factory import ClassFactory
from .context import Context
from .cortex import Cortex
from .cortex_manager import CortexManager
from .enums import ErrorCode, ErrorMsg, StateKey, Status
from .exception import CortexException
from .node import Node
from .processor import Processor, node
from .state import State

__all__ = [
    "ClassFactory",
    "Context",
    "Cortex",
    "CortexManager",
    "CortexException",
    "ErrorCode",
    "ErrorMsg",
    "Node",
    "Processor",
    "node",
    "State",
    "StateKey",
    "Status",
]



from .enums import Status


class CortexException(Exception):
    def __init__(self, status=Status.RUN_TIME_ERROR, message=None):
        self.status: Status = status
        self.message = message

        if self.message is None:
            self.message = self.status.name

        super().__init__(self.message)



import logging
from abc import ABC, abstractmethod
from typing import Callable

import yaml

from .context import Context
from .state import State


class Processor(ABC):
    trigger: Callable[[list], bool]
    stop: Callable[[str], bool]

    def __init__(self, name: str = ""):
        self.name = name
        self.class_name = ""
        self.cortex = ""
        self.group = ""
        self.events = []
        self.conditions = {}
        self.reentrant = False
        self.in_degree = 0
        self.config = {}

    def init(self) -> bool:
        return True

    @abstractmethod
    async def process(self, ctx: Context, state: State) -> bool:
        raise NotImplementedError("process method is not implemented error!")

    @property
    def is_required(self):
        return self.config.get("required", False)

    def increment_in_degree(self):
        self.in_degree += 1


def node(config: str = ""):
    def decorator(c):
        init_original = c.init

        def init_proxy(self):
            if config:
                try:
                    with open(config, "r") as f:
                        self.config = yaml.load(f, Loader=yaml.FullLoader)
                except Exception as e:
                    logging.error(f"failed to read config file: {config}, error: {e}")
                    return False

            init_original(self)

        c.init = init_proxy
        return c

    return decorator



# -*- coding: utf-8 -*-
import time

from .enums import StateKey

class State:
    def __init__(self):
        self._data = {}

    def set(self, key: str, value) -> None:
        self._data[key] = value

    def get(self, key: str, default=None):
        return self._data.get(key, default)

    def has(self, key: str):
        return key in self._data

    def append(self, key: str, value):
        self._data.setdefault(key, []).append(value)

    def timer_start(self, timer_key: str):
        self._data.setdefault(StateKey.TIME_DURATION.value, {})[timer_key] = time.time() * 1000

    def timer_end(self, timer_key: str) -> bool:
        timer_data = self._data.get(StateKey.TIME_DURATION.value, {})
        if timer_key in timer_data:
            timer_data[timer_key] = round(time.time() * 1000 - timer_data[timer_key], 2)
            return True
        return False

    def time_cost(self, timer_key) -> float:
        return round(self._data.get(StateKey.TIME_DURATION.value, {}).get(timer_key, 0), 2)

    def to_dict(self) -> dict:
        return {
            "data": self._data.copy()
        }

    def from_dict(self, data: dict) -> None:
        if data:
            self._data.update(data.get("data", {}))


import time
import unittest

from gyrus.context import Context
from gyrus.cortex_manager import CortexManager
from gyrus.state import State


class TestContext(unittest.IsolatedAsyncioTestCase):
    async def test_context(self):
        state = State()
        state.set("start_at", time.time())

        cm = CortexManager([{"name": "context", "path": "tests/cortices/context.yaml"}])
        cx = cm.get("context")

        with Context() as ctx:
            ctx.object_in_context = "this is value in context"
            success = await cx.run(ctx, state)

            self.assertEqual(state.get("context_from_case_b"), 2)
            self.assertEqual(state.get("context_from_case_c"), 2)
            self.assertEqual(state.get("object_in_context"), "this is value in context")
            self.assertEqual(success, True)


if __name__ == "__main__":
    unittest.main()



import asyncio
import time
import unittest

from gyrus.context import Context
from gyrus.cortex_manager import CortexManager
from gyrus.state import State


class TestStream(unittest.IsolatedAsyncioTestCase):
    async def test_context(self):
        state = State()
        state.set("start_at", time.time())

        cm = CortexManager([{"name": "stream", "path": "tests/cortices/stream.yaml"}])
        cx = cm.get("stream")

        messages = []

        with Context() as ctx:
            asyncio.create_task(cx.run(ctx, state))

            async for message in ctx.stream:
                messages.append(message)

        self.assertEqual(len(messages), 6)


if __name__ == "__main__":
    unittest.main()



import time
import unittest

from gyrus.context import Context
from gyrus.cortex_manager import CortexManager
from gyrus.state import State


class TestTrigger(unittest.IsolatedAsyncioTestCase):
    async def test_trigger(self):
        state = State()
        state.set("start_at", time.time())

        cm = CortexManager([{"name": "trigger", "path": "tests/cortices/trigger.yaml"}])
        cx = cm.get("trigger")

        with Context() as ctx:
            success = await cx.run(ctx, state)

            self.assertEqual(state.get("b"), 3)
            self.assertEqual(state.get("c"), 6)
            self.assertEqual(state.get("d"), 6)
            self.assertEqual(success, True)

    async def test_trigger_with_reentrant(self):
        state = State()
        state.set("start_at", time.time())
        state.set("reentrant_times", 0)

        cm = CortexManager(
            [
                {
                    "name": "trigger_reentrant",
                    "path": "tests/cortices/trigger_reentrant.yaml",
                }
            ]
        )
        cx = cm.get("trigger_reentrant")

        with Context() as ctx:
            success = await cx.run(ctx, state)

            self.assertEqual(state.get("b"), 3)
            self.assertEqual(state.get("c"), 6)
            self.assertEqual(state.get("d"), 6)
            self.assertGreaterEqual(state.get("reentrant_times"), 2)
            self.assertEqual(success, True)


if __name__ == "__main__":
    unittest.main()



import time
import unittest

from gyrus.context import Context
from gyrus.cortex_manager import CortexManager
from gyrus.enums import Status
from gyrus.exception import CortexException
from gyrus.state import State


class TestCortexManager(unittest.IsolatedAsyncioTestCase):
    async def test_circle(self):
        with self.assertRaises(CortexException) as cm:
            cm = CortexManager(
                [{"name": "circle", "path": "tests/cortices/circle.yaml"}]
            )

        self.assertEqual(cm.exception.status, Status.CORTEX_HAS_CYCLE)

    async def test_node_missing(self):
        with self.assertRaises(CortexException) as cm:
            cm = CortexManager(
                [{"name": "node_missing", "path": "tests/cortices/node_missing.yaml"}]
            )

        self.assertEqual(cm.exception.status, Status.CORTEX_NODE_MISSING)

    async def test_complex_dag(self):
        state = State()
        state.set("start_at", time.time())

        cm = CortexManager(
            [{"name": "complex_dag", "path": "tests/cortices/complex_dag.yaml"}]
        )
        cx = cm.get("complex_dag")

        with Context() as ctx:
            success = await cx.run(ctx, state)

            self.assertEqual(success, True)
            self.assertEqual(state.get("b"), 3)
            self.assertEqual(state.get("c"), 3)
            self.assertEqual(state.get("d"), 6)
            self.assertEqual(state.get("e"), 9)

    async def test_condition_simple(self):
        state = State()
        state.set("start_at", time.time())
        state.set("condition_a", "nerdduan")
        state.set("condition_b", "nerddeng")

        cm = CortexManager(
            [
                {
                    "name": "condition_simple",
                    "path": "tests/cortices/condition_simple.yaml",
                }
            ]
        )
        cx = cm.get("condition_simple")

        with Context() as ctx:
            success = await cx.run(ctx, state)

            self.assertEqual(state.get("b"), 3)
            self.assertEqual(state.get("c"), None)
            self.assertEqual(state.get("d"), 3)
            self.assertEqual(state.get("e"), 3)
            self.assertEqual(state.get("f"), 3)
            self.assertEqual(success, True)

    async def test_condition_nested(self):
        state = State()
        state.set("start_at", time.time())
        state.set("condition_a", "nerdduan")
        state.set("condition_b", "nerddeng")

        cm = CortexManager(
            [
                {
                    "name": "condition_nested",
                    "path": "tests/cortices/condition_nested.yaml",
                }
            ]
        )
        cx = cm.get("condition_nested")

        with Context() as ctx:
            success = await cx.run(ctx, state)

            self.assertEqual(state.get("b"), None)
            self.assertEqual(state.get("c"), 3)
            self.assertEqual(state.get("d"), 3)
            self.assertEqual(state.get("e"), 3)
            self.assertEqual(success, True)


if __name__ == "__main__":
    unittest.main()



import asyncio
from typing import Dict, Optional

from .context import Context
from .enums import Status
from .exception import CortexException
from .node import Node
from .processor import Processor
from .state import State


class Runtime(object):
    def __init__(self):
        self.waiting: Dict[str, Node] = {}
        self.running: Dict[str, asyncio.Task] = {}
        self.completed: Dict[str, Status] = {}

        # When concurrent execution causes multiple processors to return and
        # trigger the same event simultaneously, this event should only be
        # executed once. The execution logic conflicts caused by concurrency
        # should be resolved by the triggered processor itself.
        #
        # eg:
        # processor_a -> trigger(["event_a"])
        # processor_b -> trigger(["event_a"])
        #
        # class Node(Processor):
        #     async def process(self, ctx: Context, state: State) -> bool:
        #         if state.get("something_from_processor_a") is True:
        #             do_something_a()
        #
        #         if state.get("something_from_processor_b") is True:
        #             do_something_b()
        # ...

        self.events: Dict[str, str] = {}

    def is_visited(self, node: Processor) -> bool:
        if node.reentrant:
            return False
        if node.name in self.waiting:
            return True
        if node.name in self.running:
            return True
        if node.name in self.completed:
            return True
        return False

    def add_waiting(self, parent: str, name: str, nodes: Dict, nexts: Dict) -> bool:
        if self.is_visited(nodes[name]):
            return False

        nodes[name].trigger = self.trigger(parent)
        nodes[name].stop = self.stop

        self.waiting[name] = Node(nodes[name], nexts[name], parent=parent)
        return True

    def add_running(
        self,
        ctx: Context,
        state: State,
        name: str,
        executor,
        nodes: Optional[Dict] = None,
        nexts: Optional[Dict] = None,
    ) -> bool:
        if name in self.running:
            raise CortexException(
                Status.RUN_TIME_ERROR, f"node({name}) has been already in running"
            )

        node = self.waiting.pop(name)

        # Check if the node can be executed, if not, mark as completed and reduce in-degree of downstream nodes.
        if not node.check(state):
            # Mark the node as completed (skipped due to condition not met)
            self.completed[name] = Status.SUCCESS
            # Reduce in-degree of downstream nodes
            if nodes is not None and nexts is not None:
                self.on_node_completed(name, nodes, nexts)
            return True

        task = asyncio.get_event_loop().create_task(
            self.executor_wrapper(executor, [ctx.clone(node.parent), node, state])
        )

        # Set the name of the task to the group of the processor.
        task.set_name(node.processor.group)

        self.running[name] = task
        return True

    def get_running(self) -> list:
        return self.running.values()

    def get_task_status(self, task) -> tuple[Status, str]:
        if task.cancelled():
            return Status.NODE_CANCELLED, "Node execution cancelled"

        exception = task.exception()
        if exception is not None:
            # Convert exception object to string representation
            exception_str = str(exception)
            if not exception_str:
                exception_str = repr(exception)
            return Status.NODE_EXECUTE_FAIL, exception_str

        if task.result() is False:
            return (
                Status.NODE_EXECUTE_FAIL,
                "Node execution failed without exception details",
            )

        return Status.SUCCESS, "Node execution successful"

    def add_completed(self, node: Processor) -> tuple[Status, str]:
        if node.name in self.completed and not node.reentrant:
            raise CortexException(
                Status.RUN_TIME_ERROR,
                f"node({node.name}) has been already in completed",
            )

        task = self.running.pop(node.name)
        status, message = self.get_task_status(task)

        self.completed[node.name] = status
        return status, message

    def on_node_completed(self, name: str, nodes: Dict, nexts: Dict) -> bool:
        if name not in nexts:
            raise CortexException(
                Status.CORTEX_CONFIG_ERROR, f"get next nodes of {name} failed"
            )

        node_nexts = nexts[name]

        for child in node_nexts:
            self.add_waiting(name, child, nodes, nexts)

        for name, node in self.waiting.items():
            if name in node_nexts:
                node.in_degree -= 1
        return True

    def has_pending_nodes(self):
        return len(self.waiting) + len(self.running) > 0

    def executor_wrapper(self, executor, args):
        try:
            return executor(*args)
        except asyncio.CancelledError as _:
            return True

    def trigger(self, parent: str):
        def currying(events: list) -> bool:
            for event in events:
                self.events[event] = parent
            return True

        return currying

    def stop(self, group: str):
        if not group:
            return

        for _, node in self.running.items():
            if node.get_name() == group:
                node.cancel()

        for name, node in self.waiting.items():
            if node.processor.group == group:
                self.waiting.pop(name)



from .stream import Stream


class Context:
    def __init__(self, parent: str = "root"):
        self._attributes = {
            "parent": parent,
            "stream": Stream(maxsize=10),
        }

    def __getattr__(self, name):
        if name in self._attributes:
            return self._attributes[name]
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name, value):
        if name == "_attributes":
            super().__setattr__(name, value)
        else:
            self._attributes[name] = value

    def __delattr__(self, name):
        if name in self._attributes:
            del self._attributes[name]
        else:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._attributes["stream"].close()
        self._attributes.clear()

        return False

    def set(self, name, value):
        self._attributes[name] = value

    def get(self, name):
        return self._attributes.get(name)

    def clone(self, parent):
        ctx = Context(parent=parent)
        ctx.stream = self.stream

        for name, value in self._attributes.items():
            if name not in ["parent", "stream"]:
                ctx.set(name, value)

        return ctx



import asyncio
import logging
import time

from .class_factory import ClassFactory
from .context import Context
from .enums import Status
from .exception import CortexException
from .node import Node
from .runtime import Runtime
from .state import State

PROCESSOR_CLASS_NAME = "Node"


class Cortex(object):
    def __init__(self):
        self.name = ""
        self.nodes = {}
        self.nexts = {}
        self.state = None
        self.factory = ClassFactory()

    def build(self, cortex_name: str, config: dict):
        for step in config:
            # Load the processor class from the specified file and register it
            # with the given name.
            self.factory.load(step["name"], step["file"], PROCESSOR_CLASS_NAME)

            node = self._create_node(step)
            if node is None:
                return False

            self._add_node_to_cortex(node, step)

        if not self._validate_cortex():
            raise CortexException(Status.CORTEX_NODE_MISSING)

        self._increment_in_degree()

        for node in self.nodes.values():
            if node.in_degree == 0:
                return True

        # If there is no node with in-degree 0, the graph has a cycle.
        raise CortexException(Status.CORTEX_HAS_CYCLE)

    async def run(self, ctx: Context, state: State) -> bool:
        try:
            runtime = Runtime()
            return await self._schedule_loop(ctx, state, runtime)
        except CortexException as e:
            ctx.status = e.status
            ctx.status_message = e.message
        except Exception as e:
            ctx.status = Status.RUN_TIME_ERROR
            raise e

        # Close the stream.
        ctx.stream.close()
        return False

    async def executor(self, ctx: Context, node: Node, state: State):
        logging.debug(f"processor begin, name:{node.processor.name}")

        state.timer_start(node.processor.name)
        if node.in_degree > 0:
            logging.error(
                f"execute node:{node.processor.name} in_degree > 0, value:{node.in_degree}"
            )
            state.timer_end(node.processor.name)
            return False

        start_at = time.time()
        process_ret = await node.processor.process(ctx, state)
        logging.info(
            f"processor:{node.processor.name} executed, result:{process_ret} cost:{time.time() - start_at}"
        )

        if node.processor.is_required and not process_ret:
            logging.error(f"node execute fail:{node.processor.name}")
            state.timer_end(node.processor.name)
            return False

        state.timer_end(node.processor.name)
        return True

    def _create_node(self, step):
        node_name = step.get("name", "")
        node_file = step.get("file", "")

        node = self.factory.get(node_name)
        if node is None:
            logging.error(f"create class failed, name:{node_name}({node_file})")
            return None

        node.name = node_name
        node.cortex = self.name
        node.group = step.get("group", "")
        node.conditions = step.get("when", {})
        node.events = step.get("listen", [])
        node.reentrant = step.get("reentrant", False)
        node.init()

        return node

    def _add_node_to_cortex(self, node, step):
        self.nodes[node.name] = node

        if node.name not in self.nexts:
            self.nexts[node.name] = set()

        for item in step.get("nexts", []):
            self.nexts[node.name].add(item)

    def _validate_cortex(self):
        for nodes_set in self.nexts.values():
            for node_name in nodes_set:
                if node_name not in self.nodes:
                    return False
        return True

    def _increment_in_degree(self):
        # Increment the in-degree of all nodes in the cortex.
        for nodes_set in self.nexts.values():
            for name in nodes_set:
                self.nodes[name].increment_in_degree()

    async def _schedule_loop(
        self, ctx: Context, state: State, runtime: Runtime
    ) -> bool:
        self._initialize_nodes(runtime)

        while runtime.has_pending_nodes():
            self._schedule_nodes(ctx, state, runtime)
            await self._execute_nodes(runtime)

        # Close the output channel.
        ctx.stream.close()
        return True

    def _initialize_nodes(self, runtime: Runtime):
        for _, node in self.nodes.items():
            if node.in_degree == 0 and len(node.events) == 0:
                runtime.add_waiting("root", node.name, self.nodes, self.nexts)

    def _schedule_nodes(self, ctx: Context, state: State, runtime: Runtime):
        nodes_to_run = [
            name for name, node in runtime.waiting.items() if node.in_degree == 0
        ]

        for name in nodes_to_run:
            runtime.add_running(ctx, state, name, self.executor, self.nodes, self.nexts)

    async def _execute_nodes(self, runtime: Runtime):
        done, _ = await asyncio.wait(
            runtime.get_running(), return_when=asyncio.FIRST_COMPLETED
        )

        # Processor the events that have been triggered.
        self._process_events(runtime)

        nodes_to_complete = [
            name for name, task in runtime.running.items() if task in done
        ]

        for name in nodes_to_complete:
            status, message = self._add_completed(name, runtime)

            if status == Status.NODE_CANCELLED:
                continue

            if status != Status.SUCCESS:
                raise CortexException(status, message)

    def _add_completed(self, name: str, runtime: Runtime) -> tuple[Status, str]:
        status, message = runtime.add_completed(self.nodes[name])

        if status in [Status.SUCCESS, Status.NODE_CANCELLED]:
            runtime.on_node_completed(name, self.nodes, self.nexts)

        return status, message

    def _process_events(self, runtime):
        for name, node in self.nodes.items():
            for event in node.events:
                if event in runtime.events:
                    parent = runtime.events[event]
                    runtime.add_waiting(parent, name, self.nodes, self.nexts)
                    continue

        runtime.events.clear()



import logging
import threading
from importlib.util import module_from_spec, spec_from_file_location


class ClassFactory:
    _lock = threading.Lock()
    _inst = None

    def __init__(self):
        self._classes = {}

    def load(self, name: str, path: str, class_name: str):
        if name in self._classes:
            return True

        spec = spec_from_file_location(name, path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

        c = getattr(module, class_name, None)
        if c is None:
            logging.error(f"class not found: {path}:{class_name}")
            return False

        self._classes[name] = c
        return True

    def get(self, name: str):
        if name in self._classes:
            return self._classes[name]()

        logging.error(f"get class faied, class not found {name}")
        return None



import asyncio
from typing import Any, AsyncIterator


class Stream:
    """A Go-like channel implementation using asyncio.Queue."""

    def __init__(self, maxsize: int = 0):
        self._queue: asyncio.Queue = asyncio.Queue(maxsize=maxsize)
        self._closed = False

    async def send(self, value: Any) -> bool:
        """Send a value to the channel (like ch <- value in Go)."""
        if self._closed:
            return False

        await self._queue.put(value)
        return True

    def send_nowait(self, value: Any) -> bool:
        """Non-blocking send."""
        if self._closed:
            return False
        self._queue.put_nowait(value)
        return True

    async def receive(self) -> tuple[Any, bool]:
        """Receive a value from channel (like value, ok := <-ch in Go)."""
        if self._closed and self._queue.empty():
            return None, False
        try:
            value = await self._queue.get()
            return value, True
        except asyncio.CancelledError:
            return None, False

    def close(self):
        """Close the channel."""
        self._closed = True

    @property
    def is_closed(self) -> bool:
        return self._closed and self._queue.empty()

    async def __aiter__(self) -> AsyncIterator[Any]:
        """Iterate over channel values (like for value := range ch in Go)."""
        while True:
            if self._closed and self._queue.empty():
                break

            try:
                yield await asyncio.wait_for(self._queue.get(), timeout=0.1)
            except asyncio.TimeoutError:
                if self._closed:
                    break
                continue



from enum import Enum, unique


@unique
class Status(Enum):
    SUCCESS = 0
    RUN_TIME_ERROR = -90000
    PARAM_ERROR = -90001
    CORTEX_NOT_FOUND = -90002
    CORTEX_HAS_CYCLE = -90003
    CORTEX_CONFIG_ERROR = -90004
    CORTEX_NODE_MISSING = -90005
    NODE_EXECUTE_FAIL = -90006
    NODE_CANCELLED = -90007


@unique
class StateKey(Enum):
    TIME_DURATION = "time_duration"


@unique
class ErrorCode(Enum):
    SUCCESS = 0
    UNKNOWN_ERROR = -1


@unique
class ErrorMsg(Enum):
    SUCCESS = "success"
    UNKNOWN_ERROR = "unknown error"



import time
import unittest

from gyrus.context import Context
from gyrus.cortex_manager import CortexManager
from gyrus.state import State


class TestCortexManager(unittest.IsolatedAsyncioTestCase):
    async def test_early_stopping(self):
        state = State()
        state.set("start_at", time.time())

        cm = CortexManager(
            [{"name": "early_stopping", "path": "tests/cortices/early_stopping.yaml"}]
        )
        cx = cm.get(
            "early_stopping",
        )

        with Context() as ctx:
            success = await cx.run(ctx, state)

            self.assertEqual(state.get("b"), None)
            self.assertEqual(state.get("c"), None)
            self.assertEqual(state.get("d"), 4)
            self.assertEqual(success, True)


if __name__ == "__main__":
    unittest.main()
